import type {
    EnrichedDataService,
    InternalProvider,
    MovieData,
    OmdbProvider,
    SearchFilters,
} from '../types';
import { areStringsEqual } from '../utils/stringHelpers';

export const createOmdbEnrichedDataService = async (
    internalProvider: InternalProvider,
    omdbProvider: OmdbProvider,
): Promise<EnrichedDataService> => {
    const internalMetadataList = await internalProvider.getAllMetadata();
    const mergedMetadataList = await Promise.all(
        internalMetadataList.map(async (internalData): Promise<MovieData> => {
            const omdbData = await omdbProvider.getMetadata(
                internalData.imdbId,
            );

            if (!omdbData) {
                throw new Error(
                    `No OMDB data found for ${internalData.imdbId}`,
                );
            }

            return {
                ...omdbData,
                ...internalData,
                ratings: [...omdbData.ratings, ...internalData.ratings],
            };
        }),
    );

    return {
        getMetadataByInternalId: (requestedId) =>
            mergedMetadataList.find(
                ({ internalId }) => internalId === requestedId,
            ),
        getMetadataByImdbId: (requestedId) =>
            mergedMetadataList.find(({ imdbId }) => imdbId === requestedId),
        getSearchResults: (filters) => {
            const nonEmptyFilters = Object.entries(filters).filter(
                ([, filterValue]) => filterValue !== undefined,
            );

            return mergedMetadataList.filter((metadata) => {
                return nonEmptyFilters.every(([fieldName, filterValue]) => {
                    const metadataValue =
                        metadata[fieldName as keyof SearchFilters];
                    if (Array.isArray(metadataValue)) {
                        return metadataValue.some((singularValue) => {
                            if (typeof singularValue === 'string') {
                                return areStringsEqual(
                                    singularValue,
                                    // filterValue here is filters[fieldName], and is not undefined (because it comes from nonEmptyFilters),
                                    // so it's guaranteed to have a matching type because of SearchFilters / GenericSearchFilters definition,
                                    // but this logic is too complicated for TS to prove it automatically.
                                    // eslint-disable-next-line @typescript-eslint/no-explicit-any, @typescript-eslint/no-unsafe-argument
                                    filterValue as any,
                                );
                            } else {
                                return singularValue === filterValue;
                            }
                        });
                    } else if (typeof metadataValue === 'string') {
                        return areStringsEqual(
                            metadataValue,
                            // see comment above
                            // eslint-disable-next-line @typescript-eslint/no-explicit-any, @typescript-eslint/no-unsafe-argument
                            filterValue as any,
                        );
                    } else {
                        return metadataValue === filterValue;
                    }
                });
            });
        },
    };
};
